/*** FILEHEADER ****************************************************************
 *
 *    FILENAME:    rc5demo.c
 *    DATE:        26.07.2011
 *    AUTHOR:      Christian Stadler
 *
 *    DESCRIPTION: RC5 demo which shows usage of RC decoder driver.
 *
 ******************************************************************************/

/*** HISTORY OF CHANGE *********************************************************
 *
 *    $Log: /pic/rc5demo/rc5demo.c $
 * 
 * 5     27.07.11 12:42 Stadler
 * - added comments
 * 
 ******************************************************************************/
 
/* make compiler case sensitive */
#pragma case

#include <rc5demo.h>
#include "types.h"


/* --- RC5 driver configuration --- */
#define RC5_DATA_PIN          PIN_B1        /* IR sensor (SFH5110-36) connected to RB1 */
#define RC5_TICKS_PER_MS      (1000/26)     /* timer increments every 25.6us, */
                                            /* i.e. around 39 ticks per millisecond */
#define RC5_GetTimer()        get_timer0()  /* timer0 shall be used for RC5 decoding */
#include "rc5.h"                            /* RC5 driver include */

/* --- macros to switch on/off LED --- */
#define Led_On()        output_high(PIN_A0)
#define Led_Off()       output_low(PIN_A0)


/*****************************************************************************/
/* Interrupt_RB                                                              */
/*                                                                           */
/* Port B change interrupt service routine.                                  */
/* Used to decode RC5 IR signal.                                             */
/*****************************************************************************/
#INT_RB
void Interrupt_RB(void)
{
    Led_On();
    
    RC5_InterruptHandler();
    clear_interrupt(INT_RB);
    
    Led_Off();
}


/*****************************************************************************/
/* main                                                                      */
/*                                                                           */
/* Configures timer0 as time base for RC5 decoding and enables port B        */
/* interrupt on change.                                                      */
/* Main loop checks if a RC5 code has been received and writes the code      */
/* to the RS232 interface.                                                   */
/*****************************************************************************/
/* NOTE: Currently it only works if PIC is reset after programming? */
void main()
{
    uint16 rc5code;

    /* FOSC/4 is timer source */
    /* FOSC = 20MHz => 5MHz, i.e. timer increment every t = 1/(FOSC/4) = 200ns */
    /* with prescaler of 128 timer will increment every 25.6us */
    setup_timer_0(RTCC_INTERNAL|RTCC_DIV_128);
    
    /* configure port B interrupt on change */
    set_tris_b(0xFF);               /* all pins on port B are input */
    input_b();                      /* read port B to clear mismatch condition */
    clear_interrupt(INT_RB);        /* clear port B interrupt flag */
    enable_interrupts(INT_RB1);     /* enable RB1 interrupt on change */
    
    /* global interrupt enable */
    enable_interrupts(GLOBAL);
    
    
    printf("**** RC5 Demo started! **** \n\r");
    
    while(TRUE)
    {
        /* check if new data is available */
        if (RC5_CodeReady())
        {
            /* get code */
            rc5code = RC5_GetCode();
            
            printf("RC5 Code=0x%04LX DevAddr=0x%02X Toggle=%d Cmd=0x%02X\n\r",
                   rc5code,
                   RC5_GetDeviceAddr(rc5code),
                   RC5_GetToggleBit(rc5code),
                   RC5_GetCmd(rc5code));
        }
        
        /* this function increments the RC5 timeout timer */
        /* NOTE: Decoding will also work without calling this function, but */
        /*       it could happen that RC5 codes are sometimes not getting */
        /*       recognized because of decoding state machine stucks due */
        /*       to erroneous RC5 signal. */
        RC5_TimeoutIncrement();
    }
}
